/*
 * Copyright 2022 Synchronous Technologies Pte Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include "export_statement.h"
#include "graph.h"
#include <nlohmann/json.hpp>
using json = nlohmann::json;

namespace zefDB {
    namespace internals {
        LIBZEF_DLL_EXPORTED Graph create_from_json(std::unordered_map<blob_index,json> blobs);
        LIBZEF_DLL_EXPORTED Graph create_from_json_fixed_layout(std::vector<json> blobs);
    }
}